"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.doCommitPushPublish = void 0;
const devopsCore_1 = require("@thatdevops/core/devopsCore");
const getPkg_1 = __importDefault(require("@thatdevops/core/pkgman/getPkg"));
const getProjectPkgs_1 = __importDefault(require("@thatdevops/core/pkgman/getProjectPkgs"));
const writePkg_1 = __importDefault(require("@thatdevops/core/pkgman/writePkg"));
const echo_1 = __importDefault(require("@thatdevops/core/utils/echo"));
const spawn_1 = __importDefault(require("@thatdevops/core/utils/spawn"));
const doNothing_1 = __importDefault(require("@thats/utils/misc/doNothing"));
const chunk_1 = __importDefault(require("lodash/chunk"));
const luxon_1 = require("luxon");
const path_1 = __importDefault(require("path"));
const doFindChangedPackages_1 = __importDefault(require("./doFindChangedPackages"));
const doPublishPackage_1 = require("./doPublishPackage");
const versionBump_1 = require("./versionBump");
function doCommitPushPublish(props) {
    return __awaiter(this, void 0, void 0, function* () {
        const { rootPath, parallel = 0 } = (0, devopsCore_1.getDevOpsCore)();
        const { force = false, commitMessage = `Publish ${luxon_1.DateTime.now().toISODate()}`, bumpType = versionBump_1.BumpType.patch, } = props !== null && props !== void 0 ? props : {};
        const { publishable } = (0, devopsCore_1.getDevOpsCore)();
        if (!publishable) {
            (0, echo_1.default)('Publishing not allowed (set the dev ops "publishable" flag');
            return;
        }
        (0, echo_1.default)(`Committing "Pre ${commitMessage}"`);
        yield (0, spawn_1.default)('git', ['add', '-A']).catch(doNothing_1.default);
        yield (0, spawn_1.default)('git', ['commit', '-a', '-m', `Pre ${commitMessage}`]).catch(doNothing_1.default);
        (0, echo_1.default)('Finding changed packages');
        let changedPkgs = force ? yield (0, getProjectPkgs_1.default)() : yield (0, doFindChangedPackages_1.default)();
        if (!changedPkgs.length) {
            (0, echo_1.default)('No changes to publish');
            return;
        }
        const effectedPkgs = yield doFindEffectedPackages(changedPkgs);
        const newVersions = doCalcNewVersions(effectedPkgs, bumpType);
        (0, echo_1.default)('Updating related dependencies');
        yield doUpdateDependencies(effectedPkgs, newVersions);
        (0, echo_1.default)('Publishing effected packages');
        if (!!parallel) {
            const chunks = (0, chunk_1.default)(effectedPkgs, parallel);
            for (const chunki of chunks) {
                yield Promise.all(chunki.map(c => (0, doPublishPackage_1.doPublishPackage)(c, newVersions[c.name])));
            }
        }
        else {
            for (const pkg of effectedPkgs) {
                yield (0, doPublishPackage_1.doPublishPackage)(pkg, newVersions[pkg.name]);
            }
        }
        (0, echo_1.default)('Updating DevOps package');
        const devopsPkg = (0, getPkg_1.default)(path_1.default.join(rootPath, 'devops'));
        yield doUpdateDependencies([devopsPkg], newVersions);
        (0, echo_1.default)(`Committing "${commitMessage}"`);
        yield (0, spawn_1.default)('git', ['add', '-A']).catch(doNothing_1.default);
        yield (0, spawn_1.default)('git', ['commit', '-a', '-m', commitMessage]).catch(doNothing_1.default);
        (0, echo_1.default)('Pushing');
        yield (0, spawn_1.default)('git', ['push']).catch(doNothing_1.default);
        (0, echo_1.default)('Setting publish dates');
        const rootPkg = (0, getPkg_1.default)(rootPath);
        rootPkg.json.devops = rootPkg.json.devops || {};
        rootPkg.json.devops.publishdates = rootPkg.json.devops.publishdates || {};
        const now = luxon_1.DateTime.now().plus({ minute: 1 }).toISO();
        for (const pkg of effectedPkgs) {
            rootPkg.json.devops.publishdates[pkg.json.name] = now;
        }
        (0, writePkg_1.default)(rootPkg);
    });
}
exports.doCommitPushPublish = doCommitPushPublish;
exports.default = doCommitPushPublish;
function doCalcNewVersions(pkgs, bumpType = versionBump_1.BumpType.patch) {
    const newVersions = {};
    for (const pkg of pkgs) {
        newVersions[pkg.name] = (0, versionBump_1.versionBump)(pkg.version, bumpType);
    }
    return newVersions;
}
function doUpdateDependencies(pkgs, newVersions) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const pkg of pkgs) {
            for (const depList of ['dependencies', 'devDependencies']) {
                const deps = pkg.json[depList];
                for (const depName in deps) {
                    if (newVersions[depName]) {
                        const oldVersion = deps[depName];
                        const newVersion = newVersions[depName];
                        if (oldVersion !== newVersion) {
                            deps[depName] = newVersion;
                        }
                    }
                }
            }
            delete pkg.json.gitHead;
            (0, writePkg_1.default)(pkg);
        }
    });
}
function doFindEffectedPackages(changedPkgs) {
    return __awaiter(this, void 0, void 0, function* () {
        const allEffected = [];
        let effected = yield doFindNewEffectedPackages(changedPkgs);
        allEffected.push(...effected);
        while (effected.length) {
            const newEffected = yield doFindNewEffectedPackages(effected);
            effected = newEffected.filter(p => !allEffected.find(p2 => p2.name === p.name));
            allEffected.push(...effected);
        }
        return allEffected;
    });
}
function doFindNewEffectedPackages(changedPkgs) {
    return __awaiter(this, void 0, void 0, function* () {
        const pkgs = yield (0, getProjectPkgs_1.default)();
        const effected = [];
        for (const changedPkg of changedPkgs) {
            effected.push(changedPkg);
            for (const pkg of pkgs) {
                if (!effected.find(p => p.name === pkg.name)) {
                    for (const depList of ['dependencies', 'devDependencies']) {
                        const deps = pkg.json[depList];
                        for (const depName in deps) {
                            if (depName === changedPkg.name) {
                                effected.push(pkg);
                            }
                        }
                    }
                }
            }
        }
        return effected;
    });
}
//# sourceMappingURL=doCommitPushPublish.js.map